/****************************************************************************
** $Id: geometry.c,v 1.13 2002/08/09 21:10:51 jet Exp $
**
** Copyright (C) 2002 Jess Thrysoee <jess@thrysoee.dk>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "xtermcontrol.h"

extern char* program_name;
extern struct controlseqs ctlseqstab[];

static void window_position(int* x, int* y);
static Window get_windowid(Display* display, Window win_root, char* title);

/* The xterm widget hierarchy looks something like:

   xterm -> vt100

   where xterm is the application widget (shell) and vt100 is the terminal
   widget. Use editres to view the hierarchy. It is the coordinates of the
   shell that xterm returns (up to patch level 166), but it does not account
   for the wm decorations (I have sent a patch to the xterm maintainer to
   remedy this), therefore we need to get coordinates from the shell's parent,
   which is usaully a window manager's frame window. */

static void
window_position(int* x, int* y)
{
   Window win_root, win_current, win_child, *children;
   Window win_parent = 0;
   unsigned int nchildren;
   Display* display;
   char* windowid;
   char title[BUFSIZ];

   display = open_display();
   win_root = DefaultRootWindow(display);

   /* The environment variable WINDOWID holds the windowid for local sessions */
   windowid = (char*) getenv("WINDOWID");
   if (windowid)
      win_current = (Window) atol(windowid);
   else {
      /* If this program is run in a remote session (e.g. rsh, ssh,...) then
       * the environment variable WINDOWID is not present, so we walk the
       * widget tree in search of the xterm window title and use that to find
       * the windowid */
      get_title(title, sizeof(title), 0, ctlseqstab[GET_TITLE].ctl1);
      win_current = get_windowid(display, win_root, title);
      if (!win_current) {
         fprintf(stderr, "%s: failed to get WINDOWID\n", program_name);
         do_exit(EXIT_FAILURE);
      }
   }

   /* find the parent which is child of root */
   do {
      if (win_parent)
         win_current = win_parent;
      XQueryTree(display,
            win_current,
            &win_root,
            &win_parent,
            &children,
            &nchildren);
      XFree(children);
   } while (win_root != win_parent);
   
   XTranslateCoordinates(display,
         win_current,
         win_root,
         0, 0,
         x, y,
         &win_child);

   XFree(display);
}

Display*
open_display(void)
{
   Display* display;

   display  = XOpenDisplay(NULL);
   if (!display) {
      fprintf(stderr, "%s: option `%s': can't open display\n",
            program_name, ctlseqstab[GET_GEOMETRY].synopsis+2);
      do_exit(EXIT_FAILURE);
   }
   return display;
}

/* inspired by xwininfo */
Window
get_windowid(Display* display, Window win_root, char* title)
{
   unsigned int i, nchildren;
   Window *children, dummy;
   Window win_xterm = 0;
   char* window_name;

   if (XFetchName(display, win_root, &window_name))
      if (!strcmp(window_name, title))
            return win_root;

   if (!XQueryTree(display,
            win_root,
            &dummy,
            &dummy,
            &children,
            &nchildren))
      return 0;

   for (i = 0; i < nchildren; i++) {
      win_xterm = get_windowid(display, children[i], title);
      if (win_xterm)
         break;
   }
   XFree(children);
   XFree(window_name);

   return win_xterm;
}

void
set_geometry(int ctl1, int ctl2, char* geometry)
{
   int w, h, x, y;
   int set_pos, set_size;

   set_pos = set_size = 1;
   if (sscanf(geometry, "%d%*[xX]%d%d%d", &w, &h, &x, &y) != 4) {
      if (sscanf(geometry, "%d%d", &x, &y) != 2) {
         set_pos = 0;
         if (sscanf(geometry, "%d%*[xX]%d", &w, &h) != 2)
            set_size = 0;
      }
   }

   if (set_size)
      csi_print3(ctl1, h, w);
   
   /* negative positions are truncated to origo */
   if (set_pos) {
      x = x < 0 ? 0 : x;
      y = y < 0 ? 0 : y;
      csi_print3(ctl2, x, y);
   }
}

void
get_geometry(char* geometry, size_t size, int verbose, int ctl1, int ctl2)
{
   int w, h, x, y;
   char temp[20];

   csi_print1(ctl1);
   tty_read(temp, sizeof(temp), ';');
   if (sscanf(temp, "%*[^;];%d;%dt", &h, &w) != 2) {
      fprintf(stderr, "%s: failed to read xterm size\n", program_name);
      do_exit(EXIT_FAILURE);
   }
   window_position(&x, &y);

   if (!verbose)
      snprintf(geometry, size, "%dx%d%+d%+d", w, h, x, y);
   else
      snprintf(geometry, size, "%s=\"%dx%d%+d%+d\"",
            ctlseqstab[GET_GEOMETRY].conf_title, w, h, x, y);
}

void
validate_geometry(char* geometry)
{
   int w, h, x, y;

   w = h = 0;
   if (sscanf(geometry, "%d%*[xX]%d%d%d", &w, &h, &x, &y) != 4) {
      w = h = 0;
      if (sscanf(geometry, "%d%d", &x, &y) != 2) {
         w = h = 0;
         if (sscanf(geometry, "%d%*[xX]%d", &w, &h) != 2)
            usage(EXIT_FAILURE);
      }
   }
   
   if (w < 0 || h < 0) {
      fprintf(stderr, "%s: option `--geometry' requires non-negative width \
and height\n", program_name);
      usage(EXIT_FAILURE);
   }
}
